import gzip
import inspect
import logging
import sys
from importlib import import_module
from typing import Callable

import typer
import uvicorn
from fastapi import FastAPI, Request, Response
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.routing import APIRoute
from fastapi_migrate import Migrate
from fastapi_pagination import add_pagination
from starlette.middleware.cors import CORSMiddleware

from ._utils import singleton
from .cli import entry
from .middlewares import process_middleware
from .servicer import get_servicer, make_grpc_serve
from .utils import sync_exec

logger = logging.getLogger('bali')


class GzipRequest(Request):
    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            body = await super().body()
            if "gzip" in self.headers.getlist("Content-Encoding"):
                body = gzip.decompress(body)
            self._body = body
        return self._body


class GzipRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            request = GzipRequest(request.scope, request.receive)
            return await original_route_handler(request)

        return custom_route_handler


@singleton
class Bali:
    def __init__(self, base_settings=None, proto_dir="protos", **kwargs):
        """
        :param base_settings: base settings
        :param proto_dir: protobuf definition directory,
                            eg: protos='proto.api.v1'
        :param kwargs: other settings
        """
        self.base_settings = base_settings or {}
        self.kwargs = kwargs
        self.title = kwargs.get('title', 'Bali')

        self._app = None

        self._pb2 = None
        self._pb2_grpc = None
        self._rpc_servicer = None
        self._proto_dir = proto_dir  # default protobuf definition directory

        # Create FastAPI instance ref to `self._app`
        self.http()

        # Construct a ServiceServicer class ref to `self._rpc_servicer`
        self.rpc()

        self.resolve_declarative()

    def __getattribute__(self, attr, *args, **kwargs):
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            return getattr(self._app, attr)

    async def __call__(self, scope, receive, send) -> None:
        await self._app.__call__(scope, receive, send)  # pragma: no cover

    @property
    def pb2(self):
        return self._pb2

    @property
    def pb2_grpc(self):
        return self._pb2_grpc

    @property
    def rpc_servicer(self):
        return self._rpc_servicer

    @staticmethod
    def _launch_http():
        uvicorn.run(
            "main:app",
            port=8000,
            reload=True,
            access_log=True,
            reload_excludes=['*.log'],
        )

    async def _launch_rpc(self):
        """Launch RPC process

        `rpc_service` is the module contains gRPC components
            (a ServiceServicer class and a serve function,
            both of them is optional).
            can be a class or function, when it's a function,
            it will be called directly.

            ServiceServicer class example:

            ```python
            class FooService(pb2_grpc.FooServiceServicer):
                def GetFoo(self, request, context):
                    schema = FooResource().get(request, context)
                    return json_format.ParseDict(schema.dict(), pb2.FooEntity())
            ```

            serve function example:

            ```python
            def serve():
                port = 9080
                server = grpc.server(
                    futures.ThreadPoolExecutor(max_workers=10),
                    interceptors=[ProcessInterceptor()],
                )
                pb2_grpc.add_FooServiceServicer_to_server(FooService(), server)
                server.add_insecure_port(f'[::]:{port}')
                server.start()
                server.wait_for_termination()
            ```

            ServiceServicer will compose with registered resources

        """
        service = self.kwargs.get('rpc_service')
        if service and service.serve:
            serve = service.serve
        else:
            serve = make_grpc_serve(self)

        # Start rpc using user defined service module
        # included sync and async mode
        if inspect.iscoroutinefunction(serve):
            await serve()
        else:
            serve()

    def _launch_event(self):
        from .events import handle
        from bali import init_handler, __version__

        event_handler = self.kwargs.get('event_handler')
        if not event_handler:
            raise Exception('event_handler not provided')
        init_handler(event_handler)
        logger.info('Bali v%s Event started.', __version__)
        while True:
            handle()

    def settings(self, **kwargs):
        self.base_settings.update(kwargs)

    def http(self):
        """load FastAPI to instance"""
        self._app = FastAPI(title='Bali', **self.base_settings)
        self._app.router.route_class = GzipRoute

        # routers
        for router in self.kwargs.get('routers', []):
            self._app.include_router(**router)
        # cors
        backend_cors_origins = self.kwargs.get('backend_cors_origins')
        if backend_cors_origins:
            self._app.add_middleware(GZipMiddleware)
            self._app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )
            self._app.middleware('http')(process_middleware)

        add_pagination(self._app)

    def rpc(self):
        pb2_path = f'{self._proto_dir}.{self.title}_pb2'
        pb2_grpc_path = f'{self._proto_dir}.{self.title}_pb2_grpc'

        try:
            self._pb2 = import_module(pb2_path)
            self._pb2_grpc = import_module(pb2_grpc_path)

            # TODO: Should support service module and service class
            service = self.kwargs.get('rpc_service')
            service_classes = [
                f'{self.title.capitalize()}Service',
                f'{self.title.capitalize()}Servicer',
            ]
            for service_class in service_classes:
                if service and hasattr(service, service_class):
                    self._rpc_servicer = getattr(service, service_class)
                    break
            else:
                self._rpc_servicer = get_servicer(self)
        except ModuleNotFoundError:
            # No module named 'protos'
            #
            # No 'protos' won't launch gRPC service,
            # but other services not affected.
            pass

    def launch(
            self,
            http: bool = False,
            rpc: bool = False,
            event: bool = False,
            shell: bool = False,
    ):
        """Bali App entry for version < 4.0"""
        if not any([http, rpc, event, shell]):
            typer.echo(
                'Please provided service type: '
                '--http / --rpc / --event / --shell'
            )

        if http:
            self._launch_http()

        if rpc:
            sync_exec(self._launch_rpc())

        if event:
            self._launch_event()

        if shell:
            import code
            from bali import db, __version__
            banner = (
                f"Python {sys.version} on {sys.platform}\n"
                f"App: {self.title} (Framework: Bali v{__version__})"
            )
            ctx: dict = {'db': db}
            code.interact(banner=banner, local=ctx)

    def register(self, resources_cls, with_http=True, with_rpc=True):
        if not isinstance(resources_cls, list):
            resources_cls = [resources_cls]

        for resource_cls in resources_cls:
            if with_http:
                # Register HTTP service
                self._app.include_router(
                    router=resource_cls.as_router(),
                    prefix=resource_cls._http_endpoint,
                )

            if with_rpc:
                # Register RPC service
                if self._rpc_servicer:
                    resource_cls.as_servicer(self)

        if with_http:
            add_pagination(self._app)

    def resolve_declarative(self):
        """Resolve declarative APIs"""

        # Auto find API and load

        # Resolved declarative APIs to resource
        from bali.declarative import API
        self.register([factory() for factory in API.resources_factories])

    def start(self):
        # Integrated FastAPI-Migrate
        try:
            from bali import db
            from config import settings
            Migrate(
                self,
                model=db.Model,
                db_uri=settings.SQLALCHEMY_DATABASE_URI,
            )
        except ImportError:
            logger.debug("No `config.py` provide settings")

        entry(self)  # cli entry



"""
DB Connection

Expose `db` instance, bind managers to model.
"""

from sqla_wrapper import SQLAlchemy, BaseModel
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.decl_api import DeclarativeMeta

from .models import included_models
from .. import typing
from ..aio.sessions import AsyncSession
from ..exceptions import DBSetupException

# SQLA-Wrapper 4.x supported session proxy methods
# https://github.com/jpsca/sqla-wrapper/blob/v4.200628/sqla_wrapper/session_proxy.py
SQLA_WRAPPER_SESSION_PROXIES = (
    'query',
    'add',
    'add_all',
    'begin',
    'begin_nested',
    'commit',
    'delete',
    'execute',
    'expire',
    'expire_all',
    'expunge',
    'expunge_all',
    'flush',
    'invalidate',
    'is_modified',
    'merge',
    'prepare',
    'prune',
    'refresh',
    'remove',
    'rollback',
    'scalar',
)


# noinspection PyPep8Naming
class DB:
    def __init__(self):
        self._db = None  # sync engine
        self._async_engine = None  # async engine
        self.async_session = None  # async session maker

    def connect(
        self,
        database_uri,
        engine_options=None,
        session_options=None,
        **kwargs
    ):

        type_checker = TypeChecker(database_uri)

        engine_options = engine_options or {}
        if not type_checker.is_sqlite:
            engine_options.setdefault("pool_size", 5)
            engine_options.setdefault("pool_recycle", 2 * 60 * 60)

        session_options = session_options or {}
        # developers need to know when the ORM object needs to reload
        # from the db
        session_options.setdefault("expire_on_commit", False)

        # Sync mode db instance
        self._db = SQLAlchemy(
            database_uri,
            engine_options=engine_options,
            session_options=session_options,
        )
        self._db.Model = self._db.registry.generate_base(
            cls=BaseModel,
            name="Model",
            metaclass=AsyncModelDeclarativeMeta,
        )

        self._async_engine = create_async_engine(type_checker.async_uri)

        self.async_session = sessionmaker(
            self._async_engine,
            class_=AsyncSession,
            expire_on_commit=False,
        )

    def __getattribute__(self, attr, *args, **kwargs):
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            if not self._db:
                try:
                    from config import settings
                    self.connect(settings.SQLALCHEMY_DATABASE_URI)
                    if self._db:
                        return self.__getattribute__(attr)
                except ModuleNotFoundError:
                    pass

                raise DBSetupException()

            # BaseModels
            if attr in included_models:
                return included_models[attr](self)

            # Compatible legacy SQLA-Wrapper SessionProxy
            if attr in SQLA_WRAPPER_SESSION_PROXIES:
                return getattr(self._db.s, attr)

            return getattr(self._db, attr)


db: typing.DB = DB()


class TypeChecker:
    """Database URI checker

    check database type and ensure async schema
    """
    def __init__(self, database_uri):
        self.database_uri = database_uri

    @property
    def is_sqlite(self):
        return self.database_uri.startswith('sqlite')

    @property
    def is_mysql(self):
        raise NotImplementedError

    @property
    def is_postgres(self):
        raise NotImplementedError

    @property
    def async_uri(self):
        """
        Transform populate database schema to async format,
        which is used by SQLA-Wrapper
        """
        uri = self.database_uri
        database_schema_async_maps = [
            ('sqlite://', 'sqlite+aiosqlite://'),
            ('mysql+pymysql://', 'mysql+aiomysql://'),
            ('postgres://', 'postgresql+asyncpg://'),
        ]
        for sync_schema, async_schema in database_schema_async_maps:
            uri = uri.replace(sync_schema, async_schema)
        return uri


class AsyncModelDeclarativeMeta(DeclarativeMeta):
    """Make model support async using this metaclass"""
    def __getattribute__(self, attr):
        if attr in ('io', 'aio'):
            manager = super().__getattribute__(attr)
            return setup_manager(manager, self, prefix=attr)

        return super().__getattribute__(attr)

    # noinspection PyMethodParameters
    def __call__(self, *args, **kwargs):
        """
        Given model async supported

            ```python
            # async instance
            async_instance = Model(async=True)
            ```

        model method starts with prefix `async_`
        will transform to replace a copy replace the prefix

        for example, if you define a model with `async_foo`

            ```python
            class User(db.Base):
                async def foo(self):
                    return 'sync result'

                async def async_foo(self):
                    return 'async result'

            # call the method of async instance
            # it will return the "async result"
            user = User(async=True):
            assert user.foo() == 'async result'
            ```
        """

        instance = super().__call__(*args, **kwargs)
        # instance.aio is deprecated, will be removed in 3.5
        instance.aio = self.aio(instance)
        aio = kwargs.pop('aio', False)
        return instance._as_async() if aio else instance


def setup_manager(manager, model, prefix=''):
    if any([manager.db is None, manager.model is None]):
        manager = type(
            f'{prefix.upper()}{manager.__qualname__}',
            manager.__bases__,
            dict(manager.__dict__),
        )
        setattr(manager, 'db', model._db)
        setattr(manager, 'model', model)
    return manager



from fastapi_pagination import LimitOffsetParams, set_page
from fastapi_pagination.limit_offset import Page
from pydantic import BaseModel

from .._utils import parse_dict
from ..exceptions import ReturnTypeError
from ..paginate import paginate
from ..schemas import get_schema_in, UpdateRequest, CreateRequest
from ..utils import MessageToDict, ParseDict


# noinspection PyProtectedMember
def process_rpc(resource, func):
    """Process rpc actions

    :param resource: Resource instance
    :param func: Resource action process function
    :return:
    """
    request_data = MessageToDict(
        resource._request,
        including_default_value_fields=True,
        preserving_proto_field_name=True,
    )

    if func.__name__ == 'get':
        pk = resource._request.id
        result = func(resource, pk)
        result = parse_dict(result, schema=resource.schema)
        response_data = {'data': result}

    elif func.__name__ == 'list':
        schema_in = get_schema_in(func, default_by_action=True)
        result = func(resource, schema_in(**request_data))
        # Paginated the result queryset or iterable object
        if isinstance(result, BaseModel):
            raise ReturnTypeError(
                'Generic actions `list` should return a sequence'
            )
        else:
            set_page(Page)
            params = LimitOffsetParams(
                limit=request_data.get('limit') or 10,
                offset=request_data.get('offset'),
            )
            response_data = paginate(
                result,
                params=params,
                is_rpc=True,
                model_schema=resource.schema,
            )

    elif func.__name__ in ['create', 'update']:
        schema_in = get_schema_in(func, default_by_action=True)

        # if schema_in is not generic schema and has `data` key
        # parse data from `data`
        # eg : {'data': {'name': 'bali'}}
        if schema_in not in [CreateRequest, UpdateRequest]:
            request_data = request_data.get('data')

        schema = schema_in(**request_data)

        if isinstance(schema, CreateRequest):
            result = func(
                resource,
                resource.schema(**schema.data),
            )
        elif isinstance(schema, UpdateRequest):
            result = func(
                resource,
                resource.schema(**schema.data),
                pk=schema.id,
            )
        else:
            # Not generic `create` or `update` or been override
            result = func(resource, schema)

        result = parse_dict(result, schema=resource.schema)
        response_data = {'data': result}

    elif func.__name__ == 'delete':
        pk = resource._request.id
        result = func(resource, pk)
        response_data = {'result': bool(result)}

    else:
        # custom action
        schema_in = get_schema_in(func)
        result = func(resource, schema_in(**request_data))
        if not isinstance(result, dict):
            result = result.dict()
        response_data = result

    # Convert response data to gRPC response
    return ParseDict(
        response_data,
        resource._response_message(),
        ignore_unknown_fields=True
    )



"""
gRPC servicer
"""
import logging
from concurrent import futures

import grpc

logger = logging.getLogger('bali')


class ServicerNotFound(Exception):
    pass


class ServicerMethodNotFound(Exception):
    pass


def get_servicer(app):
    base_servicer = None
    servicer = f'{app.title.capitalize()}ServiceServicer'
    if hasattr(app._pb2_grpc, servicer):
        base_servicer = getattr(app._pb2_grpc, servicer)

    servicer = f'{app.title.capitalize()}Servicer'
    if hasattr(app._pb2_grpc, servicer):
        base_servicer = getattr(app._pb2_grpc, servicer)

    if base_servicer:

        class Service(base_servicer):
            pass

        return Service

    raise ServicerNotFound()


# noinspection PyProtectedMember
def make_grpc_serve(app):
    from .interceptors import ProcessInterceptor

    servicer_method = f'add_{app.title.capitalize()}ServiceServicer_to_server'
    if not hasattr(app._pb2_grpc, servicer_method):
        servicer_method = f'add_{app.title.capitalize()}Servicer_to_server'
        if not hasattr(app._pb2_grpc, servicer_method):
            raise ServicerMethodNotFound()

    # noinspection PyProtectedMember
    def serve():
        port = 9080
        server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=10),
            interceptors=[ProcessInterceptor()],
        )

        servicer = getattr(app._pb2_grpc, servicer_method)
        servicer(app._rpc_servicer(), server)
        server.add_insecure_port(f'[::]:{port}')
        server.start()
        logger.info("Service started on port: %s (env: %s)", port, 'default')
        server.wait_for_termination()

    return serve



"""
Environments

There a two style:
1. `default` style
    development, testing, staging, production
    ref: https://en.wikipedia.org/wiki/Deployment_environment#Environments
2. `abbreviation` style
    DEV, TEST, STAGING, PROD
    ref: https://medium.com/@buttertechn/qa-testing-what-is-dev-sit-uat-prod-ac97965ce4f

For historical reasons, `abbreviation` style is chosen.
You are welcome to submit a PR compatible `default` style.
"""

import logging
import os
from enum import Enum

logger = logging.getLogger('bali')


# noinspection PyArgumentList
STAGE = Enum('Stage', [
    'DEV',
    'TEST',
    'STAGING',
    'PROD',
])
env = os.environ.get('ENV', STAGE.DEV.name)

if env not in STAGE.__members__:
    logger.error('Provided ENV error: %s', env)
    env = 'DEV'

ENV = STAGE[env]



"""
Manager base classes

Sync Manager is usually bind to `io`, like User.io.create
Async Manager is usually bind to `aio`, like User.aio.create

Manager's default methods are aligned to SQLA-Wrapper 4.x
https://github.com/jpsca/sqla-wrapper/blob/v4.200628/sqla_wrapper/default_model.py
"""
import warnings

from sqlalchemy.future import select
from sqlalchemy.exc import IntegrityError, SQLAlchemyError


class Base:
    """Manager's Base class"""
    db = None
    model = None

    def __init__(self, instance):
        self.instance = instance


class Manager(Base):
    """Sync Manager is usually bind to `io`"""

    @classmethod
    def exists(cls, **attrs):
        """Returns whether an object with these attributes exists."""
        query = cls.model.query().filter_by(**attrs).exists()
        return bool(cls.db.s.query(query).scalar())

    @classmethod
    def create(cls, **attrs):
        """
        Create and persist a new record for the model, and returns it.
        """
        return cls.model(**attrs).save()

    @classmethod
    def create_or_first(cls, **attrs):
        """Tries to create a new record, and if it fails
        because already exists, return the first it founds."""
        try:
            return cls.create(**attrs)
        except IntegrityError:
            cls.db.s.rollback()
            return cls.first(**attrs)

    @classmethod
    def first(cls, **attrs):
        """Returns the first object found with these attributes."""
        return cls.query().filter_by(**attrs).first()

    @classmethod
    def first_or_error(cls, **attrs):
        """Returns the first object found with these attributes
        or raises a `ValueError` if it doesn't find one."""
        obj = cls.first(**attrs)
        if obj is None:
            raise ValueError
        return obj

    @classmethod
    def query(cls):
        return cls.db.s.query(cls.model)


class AsyncManager(Base):
    """Async model bind to aio"""
    @classmethod
    async def exists(cls, **attrs):
        async with cls.db.async_session() as session:
            stmt = select(cls.model).filter_by(**attrs)
            result = await session.execute(stmt)
            return bool(result.scalars().first())

    @classmethod
    async def create(cls, **attrs):
        return await cls.model(**attrs).aio.save()

    @classmethod
    async def create_or_first(cls, **attrs):
        """Tries to create a new record, and if it fails
        because already exists, return the first it founds."""
        async with cls.db.async_session() as session:
            try:
                return await cls.create(**attrs)
            except IntegrityError:
                session.rollback()
                return await cls.first(**attrs)

    @classmethod
    async def first(cls, **attrs):
        async with cls.db.async_session() as session:
            stmt = select(cls.model).filter_by(**attrs)
            result = await session.execute(stmt)
            return result.scalars().first()

    async def save(self):
        warnings.warn(
            '`Model.aio.save()` is deprecated and will be removed in v3.5, '
            'use instance.save() instead',
            DeprecationWarning,
        )
        async with self.db.async_session() as session:
            session.add(self.instance)
            await session.commit()
            return self.instance

    async def delete(self):
        warnings.warn(
            '`Model.aio.delete()` is deprecated and will be removed in v3.5, '
            'use instance.delete() instead',
            DeprecationWarning,
        )
        async with self.db.async_session() as session:
            session.delete(self.instance)
            await session.commit()



import re
from typing import Callable, Any

from grpc import Server
from pydantic import BaseModel


def singleton(cls):
    _instance = {}

    def inner(*args, **kwargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)
        return _instance[cls]

    def __clear__():
        _instance.clear()

        from bali.declarative import API
        # noinspection PyProtectedMember
        API._clear()

    inner.__clear__ = __clear__

    return inner


ServiceAdderNamePattern = re.compile(r"^add_.*_to_server$")


def get_service_adder(module) -> Callable[[object, Server], None]:
    namespace = vars(module)
    for i in dir(module):
        if ServiceAdderNamePattern.match(i):
            return namespace[i]


def pluralize(noun):
    """pluralize a given word

    ref:
    https://www.codespeedy.com/program-that-pluralize-a-given-word-in-python/

    :param noun:
    :return:
    """
    if re.search('[sxz]$', noun):
        return re.sub('$', 'es', noun)
    elif re.search('[^aeioudgkprt]h$', noun):
        return re.sub('$', 'es', noun)
    elif re.search('[aeiou]y$', noun):
        return re.sub('y$', 'ies', noun)
    else:
        return noun + 's'


def parse_dict(item: Any, schema: BaseModel = None):
    """Parse model instance, schema, dict to dict"""
    if isinstance(item, dict):
        return item

    # Transform model instance to schema
    if hasattr(item, '_sa_instance_state'):
        if not schema:
            raise ValueError(
                "Model instance can't parse to dict without schema"
            )
        return schema.from_orm(item).dict()

    return item.dict()



import collections

import humps
import pydantic
from pydantic import BaseModel, create_model

from ..decorators import action
from ..resources import Resource, GENERIC_ACTIONS

__all__ = ["API"]


class ResourceFactory:
    def __init__(self, resource_name):
        self.resource_name = resource_name
        self._schema = None
        self._actions = []

        # dict response values store
        #
        # for example:
        #   ```python
        #   API("Greeter").get({'hello': 'world'})
        #   ```
        # it will stored in `self._dict_response_values`
        #   ```python
        #   {
        #       "get": {'hello': 'world'}
        #   }
        #   ```
        self._dict_response_values = {}

    def __call__(self, *args, **kwargs):
        @action()
        def list(self, *args, **kwargs):
            if "list" in self._factory._dict_response_values:
                return self._factory._dict_response_values["list"]

            return self._factory._schema.find()

        @action()
        def create(self, *args, **kwargs):
            if "create" in self._factory._dict_response_values:
                return self._factory._dict_response_values["create"]

            raise NotImplementedError

        @action()
        def get(self, pk):
            if "get" in self._factory._dict_response_values:
                return self._factory._dict_response_values["get"]

            return self._factory._schema.find(pk)

        @action()
        def update(self, *args, **kwargs):
            if "update" in self._factory._dict_response_values:
                return self._factory._dict_response_values["update"]

            raise NotImplementedError

        @action()
        def delete(self, *args, **kwargs):
            if "delete" in self._factory._dict_response_values:
                return self._factory._dict_response_values["delete"]

            raise NotImplementedError

        type_dict = self._get_resource_type_dict(locals())

        # noinspection PyUnresolvedReferences
        return type(
            f"{humps.pascalize(self.resource_name)}Resource",
            (Resource, ),
            type_dict,
        )

    def _push_action(self, actions):
        if not isinstance(actions, list):
            actions = [actions]
        for action_name in actions:
            if action_name not in self._actions:
                self._actions.append(action_name)

    def _get_resource_type_dict(self, contexts):
        # defined resource class members
        type_dict = {
            "_factory": self,
            "schema": self._schema,
        }
        for action_name in self._actions:
            type_dict[action_name] = contexts[action_name]

        return type_dict

    def schema(self, *args, **kwargs):
        """Declare schema

        if first argument is Pydantic, means schema is the Pydantic model

        :param args:
        :param kwargs:
        :return:
        """
        if args and isinstance(args[0], pydantic.main.ModelMetaclass):
            self._schema = args[0]
        return self

    def all(self, *args, **kwargs):
        self._push_action(GENERIC_ACTIONS)
        return self

    def list(self, *args, **kwargs):
        response = args[0]

        if isinstance(response, list):
            response = response[0]

        if isinstance(response, dict):
            # noinspection PyUnresolvedReferences
            schema_name = f"{humps.pascalize(self.resource_name)}Schema"
            self._schema = create_model(schema_name, **response)

            self._dict_response_values.update(list=args[0])

        self._push_action("list")

        return self

    def create(self, func):
        self._push_action("create")
        return self

    def get(self, *args, **kwargs):
        response = args[0]

        if isinstance(response, list):
            response = response[0]

        if isinstance(response, dict):
            # noinspection PyUnresolvedReferences
            schema_name = f"{humps.pascalize(self.resource_name)}Schema"
            self._schema = create_model(schema_name, **response)

            self._dict_response_values.update(get=args[0])

        self._push_action("get")
        return self

    def update(self, func):
        self._push_action("update")
        return self

    def delete(self, func):
        self._push_action("delete")
        return self


class ModelSchema(BaseModel):
    """Declarative model schema

    Expected support in-memory/SQLAlchemy/Mongo backends

    """

    _storage = collections.OrderedDict()  # in-memory storage

    @classmethod
    def push(cls, id=None, **kwargs):
        """When condition provided, execute update, otherwise create"""
        if not id:
            if len(cls._storage) <= 0:
                next_id = 1
            else:
                next_id = cls._storage.keys()[-1] + 1
            kwargs.update(id=next_id)
            cls._storage[next_id] = kwargs

        kwargs.update(id=id)
        cls._storage[id] = kwargs
        return kwargs

    @classmethod
    def pop(cls, id):
        del cls._storage[id]
        return id

    @classmethod
    def find(cls, id=None):
        if not id:
            return [v for k, v in cls._storage.items()][:50]
        return cls._storage[id]


class DeclarativeAPI:
    def __init__(self):
        self.resources_factories = []
        self.schema = ModelSchema

    def __call__(self, *args, **kwargs):
        return self.define(args[0])

    def define(self, resource_name):
        resource_factory = ResourceFactory(resource_name)
        self.resources_factories.append(resource_factory)
        return resource_factory

    def _clear(self):
        """Reset API values, most used in unittests"""
        self.__init__()


API = DeclarativeAPI()



import typer

typer_app = typer.Typer()


def callback():
    """
    Bali App entry 4.0 style
    """


def entry(application):
    """Bali application CLI entry
    """

    # When migrate enabled, cli group will changed
    # in <4.0, ENABLE_MIGRATE default is `False`
    # in >=4.0, ENABLE_MIGRATE default is `True`
    #
    # migrate disabled: cli used as same as bali < 4.0
    #   eg: python main.py --http
    # migrate enabled: cli used as same as bali >= 4.0
    #   eg: bali run http
    try:
        from config import settings
        enable_migrate = settings.ENABLE_MIGRATE
    except:
        enable_migrate = False

    if enable_migrate:
        typer_app.command(name='run')(application.launch)
        typer_app.callback()(callback)

        from fastapi_migrate.cli import db
        typer_click_object = typer.main.get_command(typer_app)
        typer_click_object.add_command(db, "db")

        typer_click_object()
    else:
        typer_app.command()(application.launch)
        typer_app()



from datetime import datetime
from typing import Dict, Optional, Any, Type, TypeVar, Union, List, Tuple

from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import Session
from sqlalchemy.schema import Column
from sqla_wrapper import SQLAlchemy

_M = TypeVar("_M")
_DateTimeField = Union[datetime, Column[Optional[datetime]]]
_BooleanField = Union[bool, Column[Optional[bool]]]


class Base:
    __abstract__ = True
    __asdict_include_hybrid_properties__ = False

    def _asdict(self, **kwargs) -> Dict[str, Any]:
        ...

    def to_dict(self, **kwargs) -> Dict[str, Any]:
        ...

    def dict(self, **kwargs) -> Dict[str, Any]:
        ...

    @classmethod
    def exists(cls: Type[_M], **attrs) -> bool:
        ...

    @classmethod
    def create(cls: Type[_M], **attrs) -> _M:
        ...

    @classmethod
    def create_or_first(cls: Type[_M], **attrs) -> _M:
        ...

    @classmethod
    def first(cls: Type[_M], **attrs) -> Optional[_M]:
        ...

    @classmethod
    def first_or_error(cls: Type[_M], **attrs) -> _M:
        """
        raise: ValueError
        """
        ...

    @classmethod
    def query(cls: Type[_M]) -> Query:
        ...

    def save(self) -> _M:
        ...

    def delete(self) -> None:
        ...

    @classmethod
    def count(cls: Type[_M], **attrs) -> int:
        ...

    @classmethod
    def get_fields(cls: Type[_M]) -> List[str]:
        ...

    @classmethod
    def get_or_create(cls: Type[_M], defaults: Dict[str, Any],
                      **attrs) -> Tuple[_M, bool]:
        ...

    @classmethod
    def update_or_create(cls: Type[_M], defaults: Dict[str, Any],
                         **attrs) -> Tuple[_M, bool]:
        ...


class BaseModel:
    __abstract__ = True
    __asdict_include_hybrid_properties__ = False
    created_time: _DateTimeField
    updated_time: _DateTimeField
    is_active: _BooleanField

    def _asdict(self, **kwargs) -> Dict[str, Any]:
        ...

    def to_dict(self, **kwargs) -> Dict[str, Any]:
        ...

    def dict(self, **kwargs) -> Dict[str, Any]:
        ...

    @classmethod
    def exists(cls: Type[_M], **attrs) -> bool:
        ...

    @classmethod
    def create(cls: Type[_M], **attrs) -> _M:
        ...

    @classmethod
    def create_or_first(cls: Type[_M], **attrs) -> _M:
        ...

    @classmethod
    def first(cls: Type[_M], **attrs) -> Optional[_M]:
        ...

    @classmethod
    def first_or_error(cls: Type[_M], **attrs) -> _M:
        """
        raise: ValueError
        """
        ...

    @classmethod
    def query(cls: Type[_M]) -> Query:
        ...

    def save(self) -> _M:
        ...

    def delete(self) -> None:
        ...

    @classmethod
    def count(cls: Type[_M], **attrs) -> int:
        ...

    @classmethod
    def get_fields(cls: Type[_M]) -> List[str]:
        ...

    @classmethod
    def get_or_create(cls: Type[_M], defaults: Dict[str, Any],
                      **attrs) -> Tuple[_M, bool]:
        ...

    @classmethod
    def update_or_create(cls: Type[_M], defaults: Dict[str, Any],
                         **attrs) -> Tuple[_M, bool]:
        ...


class DB(SQLAlchemy):
    Base: Base
    BaseModel: BaseModel

    _async_engine: Any

    def connect(self, database_uri: str) -> None:
        ...

    @property
    def session(self) -> Session:
        ...

    def transaction(self) -> None:
        ...

    async def async_session(self) -> Any:
        ...



from typing import Container, Optional, Type

from fastapi.dependencies.utils import get_typed_signature
from pydantic import BaseConfig, BaseModel, create_model
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.properties import ColumnProperty

from .generic import *


class OrmConfig(BaseConfig):
    orm_mode = True


# noinspection PyDefaultArgument
def model_to_schema(
    db_model: Type,
    *,
    config: Type = OrmConfig,
    exclude: Container[str] = [],
    partial=False  # noqa
) -> Type[BaseModel]:
    mapper = inspect(db_model)
    fields = {}
    for attr in mapper.attrs:
        if isinstance(attr, ColumnProperty):
            if attr.columns:
                name = attr.key
                if name in exclude:
                    continue
                column = attr.columns[0]
                python_type: Optional[type] = None
                if hasattr(column.type, "impl"):
                    if hasattr(column.type.impl, "python_type"):
                        python_type = column.type.impl.python_type
                elif hasattr(column.type, "python_type"):
                    python_type = column.type.python_type
                assert python_type, f"Could not infer python_type for {column}"

                default = None
                if column.default is None and not column.nullable:
                    default = ...

                if partial:
                    fields[name] = (Optional[python_type], None)
                else:
                    fields[name] = (python_type, default)
    # noinspection PyTypeChecker
    pydantic_model = create_model(
        db_model.__name__,
        __config__=config,
        **fields  # noqa
    )
    return pydantic_model


def get_schema_in(func, default_by_action=False):
    """
    Inspects action function to get the schema

    if `default_by_action` is provided, It is generally to obtain a schema
    that can be used in a `generic action` method.
    (like: list/create/get/update/delete)
    """

    # Generic schemas bind to actions
    generic_schemas = {
        'list': ListRequest,
        'create': CreateRequest,
        'get': GetRequest,
        'update': UpdateRequest,
        'delete': DeleteRequest,
    }

    typed_signature = get_typed_signature(func)
    signature_params = typed_signature.parameters

    # 1st argument is self
    # 2st argument is schema_in
    index = 0
    for param_name, param in signature_params.items():
        index += 1
        if index == 2 or param_name == 'schema_in':
            schema_in = param.annotation
            if not schema_in:

                if default_by_action:
                    return generic_schemas.get(func.__name__)

                raise ValueError(
                    'Custom actions must provide `schema_in` argument with annotation'
                )

            return schema_in
    else:
        raise ValueError('Custom actions arguments error')



import functools
from typing import Tuple, TypeVar

from bali.db import db
from sqlalchemy import inspect
from sqlalchemy.exc import InvalidRequestError

__all__ = ["merge_session", "remove_session", "automatic_session_property"]

T = TypeVar("T")


def merge_session(obj: T) -> Tuple[T, bool]:
    if inspect(obj).transient:
        return obj, False

    if inspect(obj).session is not None:
        return obj, False

    try:
        return db.session.merge(obj, load=False), True
    except InvalidRequestError:
        return db.session.merge(obj), True


def remove_session(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        finally:
            db.session.remove()

    return wrapper


class automatic_session_property:  # noqa
    def __init__(self, func):
        self._func = func
        self._copy_func_info()

    def _copy_func_info(self):
        if self._func is None:
            return

        for member_name in [
            "__doc__",
            "__name__",
            "__module__",
        ]:
            value = getattr(self._func, member_name)
            setattr(self, member_name, value)

    def __get__(self, instance, _):
        if instance is None:
            return self

        this, is_merged = merge_session(instance)
        try:
            return self._func(this)
        finally:
            is_merged and db.session.remove()

    def __call__(self, func):
        return self.__get__(func, None)



import calendar
import os
from datetime import datetime, date, timedelta, time
from typing import Union
from typing_extensions import Literal

import pytz
from dateutil.relativedelta import relativedelta

TzInfoType = Union[type(pytz.utc), pytz.tzinfo.DstTzInfo]
StrTzInfoType = Union[TzInfoType, str]
DEFAULT_TZ_INFO = "Asia/Jakarta"
NotAwareDescription = "expects an aware datetime"


def get_current_timezone() -> TzInfoType:
    tz_info = os.environ.get("TZ", DEFAULT_TZ_INFO)
    return pytz.timezone(tz_info)


def get_current_timezone_name() -> str:
    return get_current_timezone().tzname(None)


def now() -> datetime:
    return datetime.now(pytz.utc)


def is_aware(value: datetime) -> bool:
    return value.utcoffset() is not None


def is_naive(value: datetime) -> bool:
    return value.utcoffset() is None


def make_aware(
        value: datetime,
        *,
        timezone: StrTzInfoType = None,
        is_dst: bool = False,
) -> datetime:
    assert is_naive(value), "expects a naive datetime"

    if timezone is None:
        timezone = get_current_timezone()
    elif isinstance(timezone, str):
        timezone = pytz.timezone(timezone)
    else:
        pass

    return timezone.localize(value, is_dst=is_dst)


def make_naive(
        value: datetime,
        *,
        timezone: StrTzInfoType = None,
) -> datetime:
    assert is_aware(value), NotAwareDescription

    if timezone is None:
        timezone = get_current_timezone()
    elif isinstance(timezone, str):
        timezone = pytz.timezone(timezone)
    else:
        pass

    return value.astimezone(timezone).replace(tzinfo=None)


def localtime(value: datetime = None, timezone: StrTzInfoType = None) -> datetime:
    value, timezone = value or now(), timezone or get_current_timezone()
    if isinstance(timezone, str):
        timezone = pytz.timezone(timezone)

    assert is_aware(value), NotAwareDescription
    return value.astimezone(timezone)


def localdate(value: datetime = None, timezone: StrTzInfoType = None) -> date:
    return localtime(value, timezone).date()


GRANULARITY = Literal["year", "month", "week", "day"]


def start_of(
    granularity: GRANULARITY,
    value: datetime = None,
    *,
    timezone: StrTzInfoType = None,
) -> datetime:
    value = localtime(value, timezone=timezone)
    if granularity == "year":
        result = value.replace(month=1, day=1)
    elif granularity == "month":
        result = value.replace(day=1)
    elif granularity == "week":
        result = value - timedelta(
            days=calendar.weekday(value.year, value.month, value.day)
        )
    elif granularity == "day":
        result = value
    else:
        raise ValueError("Granularity must be year, month, week or day")

    return make_aware(datetime.combine(result, time.min))


def end_of(
    granularity: GRANULARITY,
    value: datetime = None,
    *,
    timezone: StrTzInfoType = None,
) -> datetime:
    value = localtime(value, timezone=timezone)
    if granularity == "year":
        result = value.replace(month=12, day=31)
    elif granularity == "month":
        result = value + relativedelta(day=1, months=1, days=-1)
    elif granularity == "week":
        result = value + timedelta(
            days=6 - calendar.weekday(value.year, value.month, value.day)
        )
    elif granularity == "day":
        result = value
    else:
        raise ValueError("Granularity must be year, month, week or day")

    return make_aware(datetime.combine(result, time.max))



"""Model included

`db.BaseModel` is the most common model.
If you don't use `db.BaseModel`, you can compose Mixins to `db.Model`

Import Mixins in your project examples:

    ```python
    from bali.db.models import GenericModelMixin, ManagerMixin
    ```

"""

from contextvars import ContextVar
from datetime import datetime
from typing import List, Dict

import pytz
from sqlalchemy import Column, DateTime, Boolean
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import reconstructor
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.functions import func
from sqlalchemy.types import TypeDecorator

from .managers import Manager, AsyncManager
from ..aio.sessions import session_is_async
from ..utils import timezone


class AwareDateTime(TypeDecorator):
    impl = DateTime

    @property
    def python_type(self):
        return datetime

    def process_result_value(self, value, _):
        if value is not None:
            value = timezone.make_aware(value, timezone=pytz.utc)

        return value

    def process_bind_param(self, value, _):
        if value is not None and timezone.is_aware(value):
            value = timezone.make_naive(value, timezone=pytz.utc)

        return value


context_auto_commit: ContextVar[bool] = ContextVar(
    'context_auto_commit', default=True
)


def async_method_generate(self, attr):
    """Generate generic async methods"""
    async def save():
        async with self._db.async_session() as session:
            session.add(self)
            await session.commit()
            return self

    async def delete():
        async with self._db.async_session() as session:
            await session.delete(self)
            await session.commit()

    async def dummy():
        raise NotImplementedError("Removed async methods' prefix `async_`")

    methods = {
        'save': save,
        'delete': delete,
        'dummy': dummy,
    }

    return methods.get(attr)


def get_base(db):
    class Base(db.Model, AsyncModelMixin, ManagerMixin):
        __abstract__ = True
        __asdict_include_hybrid_properties__ = False

        # Bind SQLA-wrapper database to model
        _db = db

        def save(self):
            """Override default model's save"""
            global context_auto_commit
            db.s.add(self)
            db.s.commit() if context_auto_commit.get() else db.s.flush()
            return self

        def delete(self):
            """Override default model's delete"""
            global context_auto_commit
            db.s.delete(self)
            db.s.commit() if context_auto_commit.get() else db.s.flush()

    return Base


def get_base_model(db):
    class BaseModel(
        db.Model, AsyncModelMixin, GenericModelMixin, ManagerMixin
    ):
        __abstract__ = True
        __asdict_include_hybrid_properties__ = False

        # Bind SQLA-wrapper database to model
        _db = db

        @classmethod
        def exists(cls, **attrs):
            """Returns whether an object with these attributes exists."""
            query = cls.query().filter_by(**attrs).exists()
            return bool(db.s.query(query).scalar())

        @classmethod
        def create(cls, **attrs):
            """
            Create and persist a new record for the model, and returns it.
            """
            return cls(**attrs).save()

        @classmethod
        def create_or_first(cls, **attrs):
            """Tries to create a new record, and if it fails
            because already exists, return the first it founds."""
            try:
                return cls.create(**attrs)
            except IntegrityError:
                db.s.rollback()
                return cls.first(**attrs)

        @classmethod
        def first(cls, **attrs):
            """Returns the first object found with these attributes."""
            return cls.query().filter_by(**attrs).first()

        @classmethod
        def first_or_error(cls, **attrs):
            """Returns the first object found with these attributes
            or raises a `ValuError` if it doesn't find one."""
            obj = cls.first(**attrs)
            if obj is None:
                raise ValueError
            return obj

        @classmethod
        def query(cls):
            return db.s.query(cls)

        def save(self):
            """Override default model's save"""
            global context_auto_commit
            db.s.add(self)
            db.s.commit() if context_auto_commit.get() else db.s.flush()
            return self

        def delete(self):
            """Override default model's delete"""
            global context_auto_commit
            db.s.delete(self)
            db.s.commit() if context_auto_commit.get() else db.s.flush()

        def _asdict(self, **kwargs):
            include_hybrid_properties = kwargs.setdefault(
                "include_hybrid_properties",
                self.__asdict_include_hybrid_properties__
            )

            output_fields = []
            for i in inspect(type(self)).all_orm_descriptors:
                if isinstance(i, InstrumentedAttribute):
                    output_fields.append(i.key)
                elif isinstance(
                    i, hybrid_property
                ) and include_hybrid_properties:
                    output_fields.append(i.__name__)

            return {i: getattr(self, i, None) for i in output_fields}

        dict = to_dict = _asdict

        @classmethod
        def count(cls, **attrs) -> int:
            return db.s.query(func.count(cls.id)).filter_by(**attrs).scalar()

        @classmethod
        def get_fields(cls) -> List[str]:
            return [c.name for c in cls.__table__.columns]

        @classmethod
        def get_or_create(cls, defaults: Dict = None, **kwargs):
            instance = db.s.query(cls).filter_by(**kwargs).one_or_none()
            if instance:
                return instance, False

            instance = cls(**{**kwargs, **(defaults or {})})  # noqa
            try:
                db.s.add(instance)
                db.s.commit()
                return instance, True
            except SQLAlchemyError:
                db.s.rollback()
                instance = db.s.query(cls).filter_by(**kwargs).one()
                return instance, False

        @classmethod
        def update_or_create(cls, defaults: Dict = None, **kwargs):
            try:
                try:
                    instance = (
                        db.s.query(cls).filter_by(
                            **kwargs
                        ).populate_existing().with_for_update().one()
                    )
                except NoResultFound:
                    instance = cls(**{**kwargs, **(defaults or {})})  # noqa
                    try:
                        db.s.add(instance)
                        db.s.commit()
                    except SQLAlchemyError:
                        db.s.rollback()
                        instance = (
                            db.s.query(cls).filter_by(
                                **kwargs
                            ).populate_existing().with_for_update().one()
                        )
                    else:
                        return instance, True
            except SQLAlchemyError:
                db.s.rollback()
                raise
            else:
                for k, v in defaults.items():
                    setattr(instance, k, v)
                db.s.add(instance)
                db.s.commit()
                db.s.refresh(instance)
                return instance, False

    return BaseModel


# expose the include models and model creator
# Out of the box base models
included_models = {
    'Base': get_base,
    'BaseModel': get_base_model,
}

# --------------------  Models Mixins  -------------------- #


class AsyncModelMixin:
    """
    Given model async supported

        ```python
        # async instance
        async_instance = Model(async=True)
        ```

    model method starts with prefix `async_`
    will transform to replace a copy replace the prefix

    for example, if you define a model with `async_foo`

        ```python
        class User(db.Base):
            async def foo(self):
                return 'sync result'

            async def async_foo(self):
                return 'async result'

        # call the method of async instance
        # it will return the "async result"
        user = User(async=True):
        assert user.foo() == 'async result'
        ```

    """
    @reconstructor
    def init_on_load(self):
        in_async_context = session_is_async.get()
        return self._as_async() if in_async_context else self

    def _as_async(self):
        """Convert model instance to async model instance

        Here are the things to do:
        1. Bind async generic methods, like `save()`, `delete()`
        2. Removed async methods' prefix `async_`
        """
        self._is_async = True

        self.save = async_method_generate(self, 'save')
        self.delete = async_method_generate(self, 'delete')

        properties = dir(self)
        for method in properties:
            if method.startswith('async_'):
                new_method = method.replace('async_', '', 1)
                setattr(self, new_method, getattr(self, method))
                setattr(self, method, async_method_generate(self, 'dummy'))

        return self


class GenericModelMixin:
    """Generic model include the following fields"""
    created_time = Column(DateTime, default=datetime.utcnow)
    updated_time = Column(
        DateTime, default=datetime.utcnow, onupdate=datetime.utcnow
    )
    is_active = Column(Boolean, default=True)


class ManagerMixin:
    """ManagerMixin

    All sync method accessed by `io`
    All async method accessed by `aio`

    Sync examples:
        ```python
        # Sync
        User.io.first()
        # Async
        await User.aio.first()
        ```
    """
    io = Manager
    aio = AsyncManager



import asyncio
import logging
from functools import wraps
from importlib import import_module
from typing import Callable

from fastapi import FastAPI

__all__ = ["add_background_tasks", "run_every"]

logger = logging.getLogger('bali')
background_tasks = set()


def add_background_tasks(app: FastAPI, module="tasks"):
    import_module(module)
    on_startup = app.on_event("startup")
    for i in background_tasks:
        logger.info("Find Task %s", i.__name__)
        on_startup(i)


def run_every(seconds: float):
    def decorator(func: Callable):
        @wraps(func)
        def wrapped():
            async def task():
                task_name = func.__name__
                is_coroutine = asyncio.iscoroutinefunction(func)

                while True:
                    logger.info("Start task %s", task_name)
                    try:
                        if is_coroutine:
                            await func()
                        else:
                            loop = asyncio.get_event_loop()
                            await loop.run_in_executor(None, func)
                    except Exception as e:
                        logger.error("Task %s raise an error: %s", task_name, repr(e))
                    else:
                        logger.info("Task %s done", task_name)
                    finally:
                        await asyncio.sleep(seconds)

            asyncio.ensure_future(task())

        background_tasks.add(wrapped)

    return decorator



from pydantic import BaseModel

from bali.core import APIRouter, Bali, db, cache, cache_memoize
from bali.db.managers import Manager, AsyncManager
from bali.decorators import event_handler, init_handler
from bali.resources import Resource, ModelResource

__version__ = '3.5.0'


class Schema(BaseModel):
    pass



class DBSetupException(Exception):
    """Raise when database not initialized"""
    def __init__(self, message='Database session not initialized'):
        self.message = message
        super().__init__(self.message)


class ReturnTypeError(Exception):
    """Resource action return type error"""
    pass


class OperatorModelError(Exception):
    """Model not provided in django-like operator"""
    pass



"""
Extend async session
"""
from contextvars import ContextVar

from sqlalchemy.ext.asyncio.session import (
    AsyncSession as AsyncioSession,
)

session_is_async: ContextVar[bool] = ContextVar(
    'session_is_async', default=False
)


class AsyncSession(AsyncioSession):
    async def __aenter__(self):
        session_is_async.set(True)
        return await super().__aenter__()

    async def __aexit__(self, type_, value, traceback):
        result = await super().__aexit__(type_, value, traceback)
        session_is_async.set(False)
        return result



import bali
from setuptools import find_packages, setup


def read(f):
    return open(f, 'r', encoding='utf-8').read()


# https://pip.pypa.io/en/latest/user_guide/#fixing-conflicting-dependencies
INSTALL_REQUIREMENTS = read('requirements.txt').splitlines()

setup(
    name='bali-core',
    version=bali.__version__,
    description='Simplify FastAPI integrate gRPC services development',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/bali-framework/bali',
    author='Josh.Yu',
    author_email='josh.yu_8@live.com',
    license='MIT',
    install_requires=INSTALL_REQUIREMENTS,
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 5 - Production/Stable',
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP",
        "Environment :: Web Environment",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(
        exclude=[
            'examples',
            'examples.*',
            'tests',
            'tests.*',
            'docker',
            'docs',
        ]
    ),
    package_data={'bali': ['db/*.pyi']},
    include_package_data=True,
    zip_safe=False,
)



"""
gRPC interceptors
"""
import grpc
import logging

from google.protobuf import json_format

from bali.core import db

from typing import Callable, Any
from grpc.aio import ServerInterceptor
from ..core import _settings

logger = logging.getLogger('bali')


class ProcessInterceptor(ServerInterceptor):
    def setup(self):
        pass

    def teardown(self):
        try:
            db.s.remove()
        except Exception:
            pass

    async def intercept_service(
        self,
        continuation: Callable,
        handler_call_details: grpc.HandlerCallDetails,
    ) -> Any:
        self.setup()
        try:
            result = await continuation(handler_call_details)
        finally:
            self.teardown()

        return result



"""Bali Resource

A Resource layer base class to handle FastAPI and gRPC requests and responses

Resource's method input is Pydantic schema

"""

from collections import OrderedDict
from typing import Optional

import humps
from fastapi import HTTPException, status
from fastapi_pagination import LimitOffsetPage
from google.protobuf import message
from pydantic import BaseModel

from .generic_routes import *
from .._utils import pluralize
from ..routing import APIRouter
from ..schemas import ResultResponse, model_to_schema

__all__ = ['GENERIC_ACTIONS', 'Resource', 'pre_process']

GENERIC_ACTIONS = [
    'list',
    'get',
    'create',
    'update',
    'delete',
]


# noinspection PyUnresolvedReferences
class ResourceMeta(type):
    @property
    def _http_endpoint(self):
        if self.http_endpoint:
            endpoint = self.http_endpoint
        else:
            # Generate endpoint from resource name
            name = self.__name__.replace('Resource', '')
            words = humps.decamelize(name).split('_')
            words[-1] = pluralize(words[-1])
            endpoint = '-'.join(words)

        if not endpoint.startswith('/'):
            endpoint = f'/{endpoint}'
        return endpoint

    def _get_rpc_object(self, action):
        """The object being manipulated"""
        name = self.__name__.replace('Resource', '')
        if action == 'list':
            name = pluralize(name)

        return name


class Resource(metaclass=ResourceMeta):
    """Base Resource

    Generic Actions is get, list, create, update, delete
    """

    # Resource's name, should automatic recognition is not provided
    name = None
    plural_name = None

    # http_endpoint is `prefix` args for FastAPI include_router
    http_endpoint = None

    schema = None
    filters = []
    permission_classes = []

    _actions = OrderedDict()

    def __init__(self, request=None, context=None, response_message=None):
        self._request = request
        self._context = context
        self._response_message = response_message

        self._is_rpc = isinstance(self._request, message.Message)
        self._is_http = not self._is_rpc

        self.auth = BaseModel()

    # noinspection PyMethodFirstArgAssignment
    @classmethod
    def as_router(cls):
        cls = pre_process(cls)
        return RouterGenerator(cls)()

    # noinspection PyMethodFirstArgAssignment
    @classmethod
    def as_servicer(cls, app):
        return ServicerGenerator(cls)(app)


def pre_process(resource_cls):
    """Preprocess Resource"""

    # Auto generate model schema
    if resource_cls.schema is None:
        if getattr(resource_cls, 'model', None) is not None:
            resource_cls.schema = model_to_schema(
                resource_cls.model,
                partial=True,
            )

    return resource_cls


# noinspection PyUnresolvedReferences
class Generator:
    @property
    def resource_name(self):
        return self.cls.__name__.replace('Resource', '')

    @property
    def primary_key(self):
        return 'id'


# noinspection PyShadowingBuiltins,PyUnresolvedReferences,PyProtectedMember
# noinspection PyShadowingNames,DuplicatedCode
class RouterGenerator(Generator):
    """
    Generator FastAPI router from Resource class
    """
    def __init__(self, cls):
        self.cls = cls
        self.router = APIRouter()
        self._ordered_filters = self._get_ordered_filters()

    def __call__(self):

        # To fixed generic get action `/item/{id}` conflict with
        # custom action `/item/hello`,
        # must make sure get action `/item/{id}` is below custom action
        actions = sorted(
            self.cls._actions.keys(), key=lambda x: x in GENERIC_ACTIONS
        )

        # noinspection PyProtectedMember
        for action in actions:
            extra = self.cls._actions.get(action)
            if not hasattr(self.cls, action):
                continue
            self.add_route(action, extra)
        return self.router

    def check_permissions(self, resource):
        for permission_class in self.cls.permission_classes:
            permission = permission_class(resource)
            if not permission.check():
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail='Permission Denied',
                )

    def _get_ordered_filters(self):
        filters = OrderedDict()
        for item in self.cls.filters:
            k = item.keys().__iter__().__next__()
            v = item.values().__iter__().__next__()
            filters[k] = v
        return filters

    def get_endpoint(
        self, action, detail=False, methods=list, schema_in_annotation=None
    ):
        """Convert Resource instance method to FastAPI endpoint"""
        resource = self.cls()
        action_func = getattr(resource, action)

        def endpoint(request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return action_func()

        async def async_endpoint(request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return await action_func()

        def endpoint_detail(pk: int, request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return action_func(pk)

        async def async_endpoint_detail(pk: int, request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return await action_func(pk)

        def endpoint_schema(
            schema_in: BaseModel = None, request: Request = None, **kwargs
        ):
            resource._request = request
            self.check_permissions(resource)
            if 'get' in methods and schema_in_annotation:
                schema_in = schema_in_annotation(**kwargs)
            return action_func(schema_in)

        async def async_endpoint_schema(
            schema_in: BaseModel = None, request: Request = None, **kwargs
        ):
            resource._request = request
            self.check_permissions(resource)
            if 'get' in methods and schema_in_annotation:
                schema_in = schema_in_annotation(**kwargs)
            return await action_func(schema_in)

        sig = inspect.signature(getattr(self.cls, action))

        route = pick_route(action_func, async_endpoint, endpoint)
        if detail:
            route = pick_route(
                action_func, async_endpoint_detail, endpoint_detail
            )
        elif 'schema_in' in sig.parameters:
            route = pick_route(
                action_func, async_endpoint_schema, endpoint_schema
            )
            params = list(sig.parameters.values())[1:]
            if 'get' in methods and schema_in_annotation:
                # Destructor the `schema_in` to Query
                for field, annotation in schema_in_annotation.__fields__.items(
                ):
                    params = params[1:]
                    params.append(
                        inspect.Parameter(
                            name=field,
                            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                            default=annotation.default,
                            annotation=annotation.type_,
                        )
                    )
                route.__signature__ = sig.replace(parameters=params)
            else:
                params = list(sig.parameters.values())[1:]

            params.append(
                inspect.Parameter(
                    name='request',
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation=Request,
                )
            )
            route.__signature__ = sig.replace(parameters=params)

        return route

    def add_route(self, action, extra):
        if action == 'list':
            self.router.add_api_route(
                '',
                list_(self),
                methods=['GET'],
                response_model=LimitOffsetPage[self.cls.schema],
                summary=f'List {self.resource_name}'
            )
        elif action == 'create':
            self.router.add_api_route(
                '',
                create_(self),
                methods=['POST'],
                response_model=self.cls.schema and Optional[self.cls.schema],
                summary=f'Create {self.resource_name}',
                status_code=status.HTTP_201_CREATED,
            )
        elif action == 'get':
            self.router.add_api_route(
                '/{%s}' % self.primary_key,
                get_(self),
                methods=['GET'],
                response_model=self.cls.schema,
                summary=f'Get {self.resource_name}',
            )
        elif action == 'update':
            self.router.add_api_route(
                '/{%s}' % self.primary_key,
                update_(self),
                methods=['PATCH'],
                response_model=self.cls.schema,
                summary=f'Update {self.resource_name}'
            )
        elif action == 'delete':
            self.router.add_api_route(
                '/{%s}' % self.primary_key,
                delete_(self),
                methods=['DELETE'],
                response_model=ResultResponse,
                summary=f'Delete {self.resource_name}'
            )
        else:
            detail = extra.get('detail')
            methods = extra.get('methods')
            schema_in_annotation = extra.get('schema_in_annotation')
            path = '/{%s}' % self.primary_key if detail else ''
            self.router.add_api_route(
                f"{path}/{action.replace('_', '-')}",
                self.get_endpoint(
                    action,
                    detail,
                    methods=methods,
                    schema_in_annotation=schema_in_annotation,
                ),
                methods=methods,
                summary=f"{action.replace('_', ' ')}"
            )


# noinspection PyProtectedMember
class ServicerGenerator(Generator):
    """
    Generator gRPC servicer from Resource class
    """
    def __init__(self, cls):
        self.cls = cls

    def __call__(self, app):
        self.app = app
        actions = sorted(
            self.cls._actions.keys(), key=lambda x: x in GENERIC_ACTIONS
        )

        # noinspection PyProtectedMember
        for action in actions:
            extra = self.cls._actions.get(action)
            if not hasattr(self.cls, action):
                continue
            self.add_servicer(action, extra)

    def add_servicer(self, action, extra):

        response_pb = None
        method = f'{action.capitalize()}{self.cls._get_rpc_object(action)}'

        if action == 'list':
            response_pb = self.app.pb2.ListResponse
        elif action == 'create':
            response_pb = self.app.pb2.ItemResponse
        elif action == 'get':
            response_pb = self.app.pb2.ItemResponse
        elif action == 'update':
            response_pb = self.app.pb2.ItemResponse
        elif action == 'delete':
            response_pb = self.app.pb2.ResultResponse

        def servicer(_, request, context):
            resource = self.cls(
                request,
                context,
                response_pb,
            )
            return getattr(resource, action)()

        setattr(self.app._rpc_servicer, method, servicer)



from typing import Dict, Any

from pydantic import BaseModel


class Event(BaseModel):
    __amqp_name__ = 'default'

    type: str
    payload: Dict[str, Any] = {}

    @property
    def amqp_name(self):
        return self.__amqp_name__
